

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "vbuf.h"




void vb_dump (Vbuf* b)
{
  int i;
  printf("fill.tot:%ld fill.seg:%d fill.pos:%d\ndrain.tot:%ld drain.seg:%d drain.pos:%d nseg:%d useg:%d\n",
	b->fill.tot, b->fill.seg, b->fill.pos, b->drain.tot, b->drain.seg, b->drain.pos, b->nseg, b->useg);
  vb_dump_layout (b);
  for (i=0; i<b->useg; ++i) {
    printf("  seg[%d]: iov_len:%d %s\n", i, b->seg[i].iov_len, (i<b->drain.seg||i>b->fill.seg) ? "" : "*");
  }
}


void vb_dump_layout(Vbuf *b)
{
  int i;
  for (i=0; i<b->nseg; ++i) {
    if (i < b->drain.seg) putchar('f');
    else
    if (i <= b->fill.seg) putchar('U');
    else
    if (i < b->useg) putchar('f');
    else
    putchar('.');
  }
  putchar('\n');
}


// this does not necessarly return the correct seg being filled.
// it should only be used when checking if a seg is currently being
// filled to. this macro will return a hypothetical "next" seg
// if the real fill seg is already filled to the limit.

#define VBunofficial_fill_seg(b) ( (b).fill.pos == (b).seg[(b).fill.seg].iov_len ? (b).fill.seg + 1 : (b).fill.seg)



void vb_clear(Vbuf *b)
{
  static Vbuf blank = { {0,-1,0},{0,0,0}, 0,0 };
  b->fill = blank.fill;
  b->drain = blank.drain;
  b->useg = 0;
}


// read (fill) into buffer from file

int vb_read (Vbuf* b, int fd, int len)
{
  int spaceleft = 0;

//printf("vb_read: "); vb_dump(b);
//printf("vb_read: "); vb_dump_layout(b);

  if (b->useg && b->fill.seg >= 0) {
    spaceleft = b->seg[b->fill.seg].iov_len - b->fill.pos;
  }

  if (spaceleft > 0) {
    // more room in last seg					(case 1)
//printf("rcase1: more room in last seg\n");
    len = read (fd, (char*)b->seg[b->fill.seg].iov_base + b->fill.pos, spaceleft);
    if (len <= 0) return len;
    b->fill.pos += len;
    b->fill.tot += len;
    return len;
  }
  else if (b->useg > b->fill.seg+1) {
//printf("rcase2: there's more preallocated tail buffere available..\n");
    // there's more preallocated tail buffere available..	(case 2)
    b->fill.seg++;
    b->fill.pos = 0;
    return vb_read (b, fd, len);	// converted to case 1
  }
  else if (b->drain.seg*4 > b->useg) {
    // more than 1/4 of preallocated buffers have been drained and are free..	(case 3)
    int freesegs = b->drain.seg;
    int usedsegs = b->useg - freesegs;
    int needed_temp = freesegs < usedsegs ? freesegs : usedsegs;
//printf("rcase3: more than 1/4 of preallocated buffers have been drained and are free..\n");

    if (needed_temp > b->nseg - b->useg) {
      // there's not enough empty headers avaialble at end of list to perform slide
      goto allocate_more_seg_headers;
    }

    if (freesegs <= usedsegs) {

      //       /b->drain.seg
      //      /           /b->useg
      //     /           /       /b->nseg
      // fffffUUUUUUUUUUU........
      // .....UUUUUUUUUUUfffff...	step 1
      // UUUUUUUUUUUfffff........	step 2

      memmove (b->seg + b->useg, b->seg, freesegs * sizeof(struct iovec));		// step 1
      memmove (b->seg, b->seg + freesegs, (b->useg+freesegs) * sizeof(struct iovec));	// step 2
    }
    else {
      // fffffffffffffUUU........
      // ...fffffffffffffUUU.....	step 1
      // UUUfffffffffffff........	step 2

      memmove (b->seg + usedsegs, b->seg, b->useg * sizeof(struct iovec));	// step 1
      memmove (b->seg, b->seg + b->useg, usedsegs * sizeof(struct iovec));	// step 2
    }

    b->fill.seg -= freesegs;
    b->drain.seg = 0;
    return vb_read (b, fd, len);	// converted to case 2
  }
  else if (b->useg < b->nseg) {
    // more segment headers are avaiable, allocate new buffer	(case 4)
    struct iovec* s1 = b->seg + b->useg;
//printf("rcase4: more segment headers are avaiable, allocate new buffer\n");
    s1->iov_base = malloc (len);
    s1->iov_len = len;
    b->fill.seg = b->useg;
    b->fill.pos = 0;
    b->useg++;
    return vb_read (b, fd, len);	// converted to case 1
  }
  else {
//printf("rcase5a\n");
  allocate_more_seg_headers:
//printf("rcase5b: no more segment headers available, extend seg vector\n");
    // no more segment headers available, extend seg vector	(case 5)
    b->nseg = b->nseg ? 2*b->nseg : 32;
    b->seg = (struct iovec*) realloc (b->seg, b->nseg * sizeof(struct iovec));
    return vb_read (b, fd, len);	// converted to case 3 or 4
  }
}


#define VBempty(b) ((b).fill.tot == (b).drain.tot)


// write (drain) from buffer to file

int vb_write (Vbuf* b, int fd, int len)
{
  struct iovec* s0 = b->seg + b->drain.seg;

  assert (!VBempty(*b));
  assert (b->useg > 0);

//printf("vb_write: "); vb_dump(b);
//printf("vb_write: "); vb_dump_layout(b);


  if (b->drain.seg == VBunofficial_fill_seg(*b)) {
//printf("wcase1: we're draing from same segment we're currently filling.\n");
    // we're draing from same segment we're currently filling.		(case 1)
    assert (b->drain.pos < b->fill.pos);
    if (len > b->fill.pos - b->drain.pos) len = b->fill.pos - b->drain.pos;
    len = write (fd, (char*)s0->iov_base + b->drain.pos, len);
    if (len <= 0) return len;
    b->drain.pos += len;
    b->drain.tot += len;
    return len;
  }
  else if (b->drain.pos > 0) {
    int avail;
//printf("wcase3a\n");
  write_single_seg_buffer:
//printf("wcase3b: partially into first segment, but segment is filled to the end\n");
    // partially into first segment, but segment is filled to the end	(case 2)
    avail = s0->iov_len - b->drain.pos;
    if (len > avail) len = avail;
    len = write (fd, (char*)s0->iov_base + b->drain.pos, len);
    if (len <= 0) return len;
    b->drain.tot += len;
    if (len == avail) {
      // entire seg buffer written, we can move to the next seg
      b->drain.seg++;
      b->drain.pos = 0;
    } else {
      b->drain.pos += len;
    }
    return len;
  }
  else {
    // we got the first segment all to ourselves, and maybe more	(case 3)
    int avail = 0;
    int i;
//printf("wcase3: we got the first segment all to ourselves, and maybe more\n");
    // never writev too many vectors, since libc crashes at 0x8000...
    for (i = b->drain.seg; i < VBunofficial_fill_seg(*b) && i < 1000; i++) {
      if (avail + b->seg[i].iov_len > len) break;
      avail += b->seg[i].iov_len;
//printf("avail: %d\n", avail);
    }
    if (i == b->drain.seg) {
      // we're operating witin the first seg buffer alone
      goto write_single_seg_buffer;
    }
    assert (avail <= len);
    len = writev (fd, b->seg + b->drain.seg, i - b->drain.seg);

    if (len <= 0) return len;
    b->drain.tot += len;

    // loop backwards thru written segs, and stop when we find the last seg written from
    for (--i; ; --i) {

      avail -= b->seg[i].iov_len;
      assert (avail >= 0);
//printf("availx: %d len=%d (after - %d) [%d]\n", avail, len, b->seg[i].iov_len, i);
      if (len >= avail) {
	// writing stopped within this seg buffer
	b->drain.seg = i;
	b->drain.pos = len - avail;
	if (b->drain.pos == b->seg[i].iov_len) {
	  b->drain.seg++;
	  b->drain.pos = 0;
	}
	return len;
      }
    }

  }

}




/*
static Vbuf b;

static int main(int argc, char** argv)
{
  int in, out;
  in = open ("/dev/zero", O_RDONLY);
  out = open ("/dev/null", O_WRONLY);

//  vb_dump(&b);

  vb_read (&b, in, 10);
  vb_read (&b, in, 15);
  vb_read (&b, in, 20);
  vb_read (&b, in, 25);
//  vb_dump(&b);

  printf ("wrote:%d\n", vb_write (&b, out, 1));
//  vb_dump(&b);

  printf ("wrote:%d\n", vb_write (&b, out, 10));
//  vb_dump(&b);

  printf ("wrote:%d\n", vb_write (&b, out, 35));
//  vb_dump(&b);


  return 0;
}

*/

